@@ -29,6 +29,10 @@
 #endif
 #include <apt-pkg/string_view.h>
 
+#ifdef APT_COMPILING_APT
+#include <xxhash.h>
+#endif
+
 class FileFd;
 class pkgSourceList;
 class OpProgress;
@@ -46,7 +50,7 @@ class APT_HIDDEN pkgCacheGenerator					/*{{{*/
    }
 
    // Dirty hack for public users that do not use C++11 yet
-#if __cplusplus >= 201103L
+#if __cplusplus >= 201103L && defined(APT_COMPILING_APT)
    struct string_pointer {
       const char *data_;
       size_t size;
@@ -63,11 +67,7 @@ class APT_HIDDEN pkgCacheGenerator					/*{{{*/
    };
    struct hash {
       uint32_t operator()(string_pointer const &that) const {
-	 uint32_t Hash = 5381;
-	 const char * const end = that.data() + that.size;
-	 for (const char *I = that.data(); I != end; ++I)
-	    Hash = 33 * Hash + *I;
-	 return Hash;
+	 return XXH3_64bits(that.data(), that.size) & 0xFFFFFFFF;
       }
    };
 
@@ -146,7 +146,7 @@ class APT_HIDDEN pkgCacheGenerator					/*{{{*/
 			MMap **OutMap,pkgCache **OutCache, bool AllowMem = false);
    APT_PUBLIC static bool MakeOnlyStatusCache(OpProgress *Progress,DynamicMMap **OutMap);
 
-   void ReMap(void const * const oldMap, void const * const newMap, size_t oldSize);
+   void ReMap(void const * const oldMap, void * const newMap, size_t oldSize);
    bool Start();
 
    pkgCacheGenerator(DynamicMMap *Map,OpProgress *Progress);
